# THIS FILE IS GENERATED FROM ENSTALLER SETUP.PY
version = '4.6.3'
full_version = '4.6.3'
git_revision = 'afadb781a1719227b9ea16cba1be0c3c61756dab'
is_released = True

if not is_released:
    version = full_version
